#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "piano.h"
#include <string.h>
#include "sndfile.h"

double Piano :: freqTable[NUM_NOTES];

void Piano :: fillFrequencyTable() {
  double NOTE_UP_SCALAR = pow(2.0,1.0/12.0); 
  double A = 6.875;	// A
  A *= NOTE_UP_SCALAR;	// A#
  A *= NOTE_UP_SCALAR;	// B
  A *= NOTE_UP_SCALAR;	// C, frequency of midi note 0
  for (int i = 0; (i < NUM_NOTES); i++)	// 128 midi notes
    {
      freqTable[i] = A;
      A *= NOTE_UP_SCALAR;
    }
}

long Piano :: go(float *out, int samples) 
{
  long n = 0;
  for(int i=0;i<samples;i++) {
    if(sample > this->samples)
      break;
    n++;
    float vstring = 0.0;
    for(int k=0;k<nstrings;k++) {
      vstring += string[k]->input_velocity();
    }
    float hload = hammer->load(t,vstring/nstrings);
    float load = 0;
    for(int k=0;k<nstrings;k++) {
      load += (2*Z*string[k]->go_hammer(hload/(2*Z)))/(Z*nstrings+Zb);
    }
    float output = 0.0;
    for(int k=0;k<nstrings;k++) {
      output += string[k]->go_soundboard(load);
    }
    
    output = soundboard->reverb(output);
    
    //output += filter(output,&shaping1);
    //output = filter(output,&shaping2);
    //output += filter(output,&shaping3);
    t += dt;
    sample++;
    out[i] = output * 100.0;
  }
  return n;
}

Piano :: Piano(int note, float Fs, float v0, long samples) {
  this->Fs = Fs;
  this->v0 = v0;
  this->samples = samples;
  sample = 0;
  t = 0.0;
  dt = 1.0/Fs;

  float f = freqTable[note];
  this->f = f;

  float f0 = 27.5;
  float rho = 7850.0;
  float p = 2.0+1.0*log(f/f0)/log(4192/f0);
  float m = .06 - .058*pow(log(f/f0)/log(4192/f0),.1);
  float K = 40/pow(.7e-3,p);
  float L = 1.4 - 1.32*log(f/f0) / log(4192.0/f0);
  L = .04 + 1.4/(1+exp(-3.4+1.4*log(f/f0)));
  float r = .002*pow(1+.6*log(f/f0),-1.4);
  float rhoL = PI*r*r*rho;
  float T = (2*L*f)*(2*L*f)*rhoL;
  Z = sqrt(T*rhoL);
  Zb = 4000.0;
  Zh = 0;
  float E = 200e9;
  float flong = sqrt(E/rho) / (2.0*L);


  float rcore = (r<.0006)?r:.0006;
  float B = (PI*PI*PI)*E*(rcore*rcore*rcore*rcore)/(4.0*L*L*T);
  float hp = 1.0/7.0;

  printf("f = %g, r = %g mm, L = %g, T = %g, hammer = %g, Z = %g, K = %g, B = %g\n",f,1000*r,L,T,hp,Z,K,B);

  float c1 = 0.25;
  float c3 = 5.85;  
  if(note<31)
    nstrings = 1;
  else if(note<41)
    nstrings = 2;
  else 
    nstrings = 3;
  
  float c1b = 20.0;
  float c3b = 20.0;
  for(int k=0;k<nstrings;k++) {
    string[k] = new dwgs(f*TUNE[k],Fs,hp,c1,c3,B,Z,Zb+(nstrings-1)*Z,Zh);
  }

  float a = -1.0/4.0;
  float mix = 1;
  float alpha = 0.1e-4*log(f/f0)/log(4192/f0);
  soundboard = new Reverb(c1b,c3b,a,mix,Fs);
  hammer = new Hammer(f,Fs,m,K,p,Z,alpha,v0);
    
  biquad(500.0,Fs,10,notch,&shaping1);
  biquad(200.0,Fs,1.0,high,&shaping2);
  biquad(800.0,Fs,1.0,low,&shaping3);

}

Piano :: ~Piano() {
  for(int k=0;k<nstrings;k++) {
    delete string[k];
  }  
  delete hammer;
  delete soundboard;
}

void usage() {
  printf("usage: piano <time(s)> <note[21...108](A4=69)> <velocity[0-10](m/s)> [-b](binary file output to wave.out instead of wav output to out.wav)\n");
  exit(-1);
}

int main(int argc, char **argv) {
  if(argc != 4 && argc != 5) usage();
  float ttot = atof(argv[1]);
  int note = atoi(argv[2]);
  if(note < 21 || note > 108) usage();
  float v0 = atof(argv[3]);
  if(v0<0 || v0>10) usage();
  int bWavOutput;
  if(argc > 4 && !strcmp(argv[4],"-b"))
    bWavOutput = 0;
  else 
    bWavOutput = 1;
  
  Piano :: fillFrequencyTable();
  int Fs = 44100;

  long samples = (int)(ttot*Fs);
  FILE *out;
  PcmWriter *pcmWriter;
  if(bWavOutput)
    pcmWriter = new PcmWriter("out.wav",samples,Fs,1);
  else
    out = fopen("wave.out","w");
  Piano *piano = new Piano(note, (float)Fs, v0, samples);

  float buf[128];
  long nout;
  while(nout = piano->go(buf,128)) {
    if(bWavOutput)
      pcmWriter->write(buf,128);
    else
      fwrite(buf,sizeof(float),128,out);
  }
  if(bWavOutput)
    delete pcmWriter;
  else
    fclose(out);  
}
